/*
 * Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file audiod_command.c Commands for para_audiod. */

#include <netinet/in.h>
#include <sys/socket.h>
#include <regex.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <netdb.h>
#include <lopsub.h>

#include "para.h"
#include "audiod_cmd.lsg.h"
#include "audiod.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "buffer_tree.h"
#include "filter.h"
#include "grab_client.h"
#include "error.h"
#include "audiod.h"
#include "net.h"
#include "daemon.h"
#include "string.h"
#include "write.h"
#include "fd.h"
#include "version.h"

extern struct sched sched;
extern char *stat_item_values[NUM_STAT_ITEMS];

/** The maximal number of simultaneous connections. */
#define MAX_STAT_CLIENTS 50

/** Pointer to a command handler function. */
typedef int (*audiod_cmd_handler_t)(int, struct lls_parse_result *);

/** The lopsub user_data pointer. Only the command handler at the moment. */
struct audiod_command_info {
	audiod_cmd_handler_t handler; /**< Implementation of the command. */
};

/** Define the user_data pointer as expected by lopsub. */
#define EXPORT_AUDIOD_CMD_HANDLER(_cmd) \
	/** Implementation of _cmd. */ \
	const struct audiod_command_info lsg_audiod_cmd_com_ ## _cmd ## _user_data = { \
		.handler = com_ ## _cmd \
	};

/** Flags used for the stat command of para_audiod. */
enum stat_client_flags {
	/** Enable parser-friendly output. */
	SCF_PARSER_FRIENDLY = 1,
};

/**
 * Describes a status client of para_audiod.
 *
 * There's one such structure per audiod client that sent the 'stat' command.
 *
 * A status client is identified by its file descriptor.  para_audiod
 * keeps a list of connected status clients.
 */
struct stat_client {
	/** The stat client's file descriptor. */
	int fd;
	/** Bitmask of those status items the client is interested in. */
	uint64_t item_mask;
	/** See \ref stat_client flags. */
	unsigned flags;
	/** Its entry in the list of stat clients. */
	struct list_head node;
};

static INITIALIZED_LIST_HEAD(client_list);
static int num_clients;

/** The list of all status items used by para_{server,audiod,gui}. */
const char *status_item_list[] = {STATUS_ITEM_ARRAY};

static void dump_stat_client_list(void)
{
	struct stat_client *sc;

	list_for_each_entry(sc, &client_list, node)
		PARA_INFO_LOG("stat client on fd %d\n", sc->fd);
}
/**
 * Add a status client to the list.
 *
 * \param fd The file descriptor of the client.
 * \param mask Bitfield of status items for this client.
 * \param parser_friendly Enable parser-friendly output mode.
 *
 * Only those status items having the bit set in \a mask will be
 * sent to the client.
 *
 * \return Positive value on success, or -E_TOO_MANY_CLIENTS if
 * the number of connected clients exceeds #MAX_STAT_CLIENTS.
 */
static int stat_client_add(int fd, uint64_t mask, int parser_friendly)
{
	struct stat_client *new_client;
	int ret;

	if (num_clients >= MAX_STAT_CLIENTS) {
		PARA_ERROR_LOG("maximal number of stat clients (%d) exceeded\n",
			MAX_STAT_CLIENTS);
		return -E_TOO_MANY_CLIENTS;
	}
	ret = dup(fd);
	if (ret < 0)
		return -ERRNO_TO_PARA_ERROR(errno);
	new_client = para_calloc(sizeof(*new_client));
	new_client->fd = ret;
	PARA_INFO_LOG("adding client on fd %d\n", new_client->fd);
	new_client->item_mask = mask;
	if (parser_friendly)
		new_client->flags = SCF_PARSER_FRIENDLY;
	para_list_add(&new_client->node, &client_list);
	dump_stat_client_list();
	num_clients++;
	return 1;
}

static void close_stat_client(struct stat_client *sc)
{
	PARA_INFO_LOG("closing client fd %d\n", sc->fd);
	close(sc->fd);
	list_del(&sc->node);
	free(sc);
	num_clients--;
}

/**
 * Empty the status clients list.
 *
 * This iterates over the list of connected status clients, closes each client
 * file descriptor and frees the resources.
 */
void close_stat_clients(void)
{
	struct stat_client *sc, *tmp;

	list_for_each_entry_safe(sc, tmp, &client_list, node)
		close_stat_client(sc);
	assert(num_clients == 0);
}

/**
 * Write a message to all connected status clients.
 *
 * \param item_num The number of the status item of \a msg.
 *
 * On write errors, remove the status client from the client list and close its
 * file descriptor.
 */
void stat_client_write_item(int item_num)
{
	struct stat_client *sc, *tmp;
	struct para_buffer pb = {.flags = 0};
	struct para_buffer pfpb = {.flags = PBF_SIZE_PREFIX};
	const uint64_t one = 1;
	char *msg = stat_item_values[item_num];
	struct para_buffer *b;

	list_for_each_entry_safe(sc, tmp, &client_list, node) {
		int ret;

		if (!((one << item_num) & sc->item_mask))
			continue;
		b = (sc->flags & SCF_PARSER_FRIENDLY)? &pfpb : &pb;
		if (!b->buf)
			WRITE_STATUS_ITEM(b, item_num, "%s\n", msg? msg : "");
		ret = write(sc->fd, b->buf, b->offset);
		if (ret == b->offset)
			continue;
		/* write error or short write */
		close_stat_client(sc);
		dump_stat_client_list();
	}
	free(pb.buf);
	free(pfpb.buf);
}

/**
 * Check if string is a known status item.
 *
 * \param item Buffer containing the text to check.
 *
 * \return If \a item is a valid status item, the number of that status item is
 * returned. Otherwise, this function returns \p -E_UNKNOWN_STAT_ITEM.
 */
static int stat_item_valid(const char *item)
{
	int i;
	if (!item || !*item) {
		PARA_ERROR_LOG("%s\n", "no item");
		return -E_UNKNOWN_STAT_ITEM;
	}
	FOR_EACH_STATUS_ITEM(i)
		if (!strcmp(status_item_list[i], item))
			return i;
	PARA_ERROR_LOG("invalid stat item: %s\n", item);
	return -E_UNKNOWN_STAT_ITEM;
}

static int client_write(int fd, const char *buf)
{
	size_t len;

	if (!buf)
		return 0;
	len = strlen(buf);
	return write(fd, buf, len) != len? -E_CLIENT_WRITE: 1;
}

__malloc static char *audiod_status_string(void)
{
	const char *status = (audiod_status == AUDIOD_ON)?
		"on" : (audiod_status == AUDIOD_OFF)? "off": "sb";
	return para_strdup(status);
}

static int dump_commands(int fd)
{
	char *buf = para_strdup(""), *tmp = NULL;
	int i;
	ssize_t ret;
	const struct lls_command *cmd;

	for (i = 1; (cmd = lls_cmd(i, audiod_cmd_suite)); i++) {
		tmp = make_message("%s%s\t%s\n", buf, lls_command_name(cmd),
			lls_purpose(cmd));
		free(buf);
		buf = tmp;
	}
	ret = client_write(fd, buf);
	free(buf);
	return ret;
}

static int com_help(int fd, struct lls_parse_result *lpr)
{
	int ret;
	char *buf, *errctx;
	const struct lls_command *cmd;

	ret = lls(lls_check_arg_count(lpr, 0, 1, &errctx));
	if (ret < 0) {
		if (errctx) {
			buf = make_message("%s\n", errctx);
			client_write(fd, buf);
			free(buf);
			free(errctx);
		}
		return ret;
	}
	if (lls_num_inputs(lpr) == 0)
		return dump_commands(fd);
	ret = lls(lls_lookup_subcmd(lls_input(0, lpr), audiod_cmd_suite,
		&errctx));
	if (ret < 0) {
		buf = make_message("%s: %s\nAvailable commands:\n", errctx?
			errctx : lls_input(0, lpr), para_strerror(-ret));
		if (client_write(fd, buf) >= 0)
			dump_commands(fd);
		free(errctx);
		free(buf);
		goto out;
	}
	cmd = lls_cmd(ret, audiod_cmd_suite);
	buf = lls_long_help(cmd);
	assert(buf);
	ret = client_write(fd, buf);
	free(buf);
out:
	return ret < 0? ret : 0;
}
EXPORT_AUDIOD_CMD_HANDLER(help)

static int com_tasks(int fd, __a_unused struct lls_parse_result *lpr)
{
	int ret;
	char *tl = get_task_list(&sched);

	if (!tl) /* no tasks registered yet */
		return 0;
	ret = client_write(fd, tl);
	free(tl);
	return ret;
}
EXPORT_AUDIOD_CMD_HANDLER(tasks)

static int com_stat(int fd, struct lls_parse_result *lpr)
{
	int i, ret, parser_friendly = 0;
	uint64_t mask = 0;
	const uint64_t one = 1;
	struct para_buffer b = {.flags = 0};
	const struct lls_opt_result *r;
	unsigned num_inputs;

	ret = mark_fd_nonblocking(fd);
	if (ret < 0)
		return ret;
	r = lls_opt_result(LSG_AUDIOD_CMD_STAT_OPT_PARSER_FRIENDLY, lpr);
	if (lls_opt_given(r) > 0) {
		parser_friendly = 1;
		b.flags = PBF_SIZE_PREFIX;
	}
	num_inputs = lls_num_inputs(lpr);
	if (num_inputs == 0)
		mask--; /* set all bits */
	for (i = 0; i < num_inputs; i++) {
		ret = stat_item_valid(lls_input(i, lpr));
		if (ret < 0)
			return ret;
		mask |= (one << ret);
	}
	PARA_INFO_LOG("mask: 0x%llx\n", (long long unsigned)mask);
	FOR_EACH_STATUS_ITEM(i) {
		char *item = stat_item_values[i];
		if (!((one << i) & mask))
			continue;
		WRITE_STATUS_ITEM(&b, i, "%s\n", item? item : "");
	}
	ret = client_write(fd, b.buf);
	if (ret >= 0)
		ret = stat_client_add(fd, mask, parser_friendly);
	free(b.buf);
	return ret < 0? ret : 0;
}
EXPORT_AUDIOD_CMD_HANDLER(stat)

static int com_grab(int fd, struct lls_parse_result *lpr)
{
	int ret = grab_client_new(fd, lpr, &sched);
	return ret < 0? ret : 0;
}
EXPORT_AUDIOD_CMD_HANDLER(grab)

static int com_term(__a_unused int fd, __a_unused struct lls_parse_result *lpr)
{
	return -E_AUDIOD_TERM;
}
EXPORT_AUDIOD_CMD_HANDLER(term)

static int com_on(__a_unused int fd, __a_unused struct lls_parse_result *lpr)
{
	audiod_status = AUDIOD_ON;
	return 1;
}
EXPORT_AUDIOD_CMD_HANDLER(on)

static int com_off(__a_unused int fd, __a_unused struct lls_parse_result *lpr)
{
	audiod_status = AUDIOD_OFF;
	return 1;
}
EXPORT_AUDIOD_CMD_HANDLER(off)

static int com_sb(__a_unused int fd, __a_unused struct lls_parse_result *lpr)
{
	audiod_status = AUDIOD_STANDBY;
	return 1;
}
EXPORT_AUDIOD_CMD_HANDLER(sb)

static int com_cycle(__a_unused int fd, __a_unused struct lls_parse_result *lpr)
{
	switch (audiod_status) {
		case  AUDIOD_ON: audiod_status = AUDIOD_STANDBY; break;
		case  AUDIOD_OFF: audiod_status = AUDIOD_ON; break;
		case  AUDIOD_STANDBY: audiod_status = AUDIOD_OFF; break;
	}
	return 1;
}
EXPORT_AUDIOD_CMD_HANDLER(cycle)

static int com_version(int fd, struct lls_parse_result *lpr)
{
	int ret;
	char *msg;
	const struct lls_opt_result *r_v;

	r_v = lls_opt_result(LSG_AUDIOD_CMD_VERSION_OPT_VERBOSE, lpr);
	if (lls_opt_given(r_v))
		msg = make_message("%s", version_text("audiod"));
	else
		msg = make_message("%s\n", version_single_line("audiod"));
	ret = client_write(fd, msg);
	free(msg);
	return ret < 0? ret : 0;
}
EXPORT_AUDIOD_CMD_HANDLER(version)

/**
 * Handle arriving connections on the local socket.
 *
 * \param accept_fd The fd to accept connections on.
 * \param rfds If \a accept_fd is not set in \a rfds, do nothing.
 *
 * This is called in each iteration of the select loop. If there is an incoming
 * connection on \a accept_fd, this function reads the command sent by the peer,
 * checks the connecting user's permissions by using unix socket credentials
 * (if supported by the OS) and calls the corresponding command handler if
 * permissions are OK.
 *
 * \return Positive on success, negative on errors, zero if there was no
 * connection to accept.
 *
 * \sa para_accept(), recv_cred_buffer()
 * */
int handle_connect(int accept_fd, fd_set *rfds)
{
	int argc, ret, clifd;
	char buf[MAXLINE], **argv = NULL;
	struct sockaddr_un unix_addr;
	uid_t uid;
	const struct lls_command *cmd;
	struct lls_parse_result *lpr;
	char *errctx;
	const struct audiod_command_info *aci;

	ret = para_accept(accept_fd, rfds, &unix_addr, sizeof(struct sockaddr_un), &clifd);
	if (ret <= 0)
		return ret;
	ret = recv_cred_buffer(clifd, buf, sizeof(buf) - 1);
	if (ret < 0)
		goto out;
	uid = ret;
	PARA_INFO_LOG("connection from UID %d, buf: %s\n", ret, buf);
	ret = -E_UCRED_PERM;
	if (!uid_is_whitelisted(uid))
		goto out;
	ret = create_argv(buf, "\n", &argv);
	if (ret <= 0)
		goto out;
	argc = ret;
	ret = lls(lls_lookup_subcmd(argv[0], audiod_cmd_suite, &errctx));
	if (ret < 0)
		goto out;
	cmd = lls_cmd(ret, audiod_cmd_suite);
	ret = lls(lls_parse(argc, argv, cmd, &lpr, &errctx));
	if (ret < 0)
		goto out;
	aci = lls_user_data(cmd);
	ret = aci->handler(clifd, lpr);
	lls_free_parse_result(lpr, cmd);
out:
	free_argv(argv);
	if (ret < 0 && ret != -E_CLIENT_WRITE) {
		char *tmp;
		if (errctx) {
			tmp = make_message("%s\n", errctx);
			free(errctx);
			client_write(clifd, tmp);
			free(tmp);
		}
		tmp = make_message("%s\n", para_strerror(-ret));
		client_write(clifd, tmp);
		free(tmp);
	}
	close(clifd);
	return ret;
}

/**
 * Send the current audiod status to all connected stat clients.
 *
 * \param force Whether to write unchanged items.
 */
void audiod_status_dump(bool force)
{
	char *old, *new;

	old = stat_item_values[SI_PLAY_TIME];
	new = get_time_string();
	if (new) {
		if (force || !old || strcmp(old, new)) {
			free(old);
			stat_item_values[SI_PLAY_TIME] = new;
			stat_client_write_item(SI_PLAY_TIME);
		} else
			free(new);
	}

	new = daemon_get_uptime_str(now);
	old = stat_item_values[SI_AUDIOD_UPTIME];
	if (force || !old || strcmp(old, new)) {
		free(old);
		stat_item_values[SI_AUDIOD_UPTIME] = new;
		stat_client_write_item(SI_AUDIOD_UPTIME);
	} else
		free(new);

	old = stat_item_values[SI_AUDIOD_STATUS];
	new = audiod_status_string();
	if (force || !old || strcmp(old, new)) {
		free(old);
		stat_item_values[SI_AUDIOD_STATUS] = new;
		stat_client_write_item(SI_AUDIOD_STATUS);
	} else
		free(new);

	old = stat_item_values[SI_DECODER_FLAGS];
	new = audiod_get_decoder_flags();
	if (force || !old || strcmp(old, new)) {
		free(old);
		stat_item_values[SI_DECODER_FLAGS] = new;
		stat_client_write_item(SI_DECODER_FLAGS);
	} else
		free(new);
}

/**
 * Flush and send all status items.
 *
 * Send to  each connected client the full status item list
 * with empty values.
 */
void clear_and_dump_items(void)
{
	int i;

	FOR_EACH_STATUS_ITEM(i) {
		free(stat_item_values[i]);
		stat_item_values[i] = NULL;
		stat_client_write_item(i);
	}
}
