/*
 * Copyright (C) 2005 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write.c Paraslash's standalone wav/raw player. */

#include <regex.h>
#include <sys/types.h>

#include "para.h"
#include "string.h"
#include "write.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "stdin.h"
#include "buffer_tree.h"
#include "write.h"
#include "write_common.h"
#include "fd.h"
#include "error.h"
#include "version.h"
#include "check_wav.h"

/** Array of error strings. */
DEFINE_PARA_ERRLIST;

static struct write_args_info conf;

static struct stdin_task sit;

static int loglevel;
INIT_STDERR_LOGGING(loglevel)

__noreturn static void print_help_and_die(void)
{
	struct ggo_help h = DEFINE_GGO_HELP(write);
	bool d = conf.detailed_help_given;

	ggo_print_help(&h, d? GPH_STANDARD_FLAGS_DETAILED : GPH_STANDARD_FLAGS);
	print_writer_helps(d? GPH_MODULE_FLAGS_DETAILED : GPH_MODULE_FLAGS);
	exit(0);
}

/*
 * Parse config and register a task for a writer node.
 *
 * \param arg Command line arguments.
 * \param parent The new node will be a child of \a parent.
 * \param wn The writer node.
 *
 * If arg is \p NULL, the OS-dependent default writer is used with no
 * arguments.  The default writers are alsa for Linux, osx for OS X, oss for
 * *BSD, and the file writer if the default writer is not supported.
 *
 * Once the writer configuration has been retrieved from the ->parse_config
 * callback a writer node is created, its buffer tree node is added to the
 * buffer tree as a child of the given parent.
 *
 * Finally, the new writer node's task structure is initialized and registered
 * to the paraslash scheduler.
 *
 * \return Standard.
 */
static void setup_writer_node(const char *arg, struct btr_node *parent,
		struct writer_node *wn, struct sched *s)
{
	wn->conf = check_writer_arg_or_die(arg, &wn->writer_num);
	register_writer_node(wn, parent, s);
}

struct write_task {
	struct task *task;
	struct check_wav_context *cwc;
};

static void write_pre_select(struct sched *s, void *context)
{
	struct write_task *wt = context;
	check_wav_pre_select(s, wt->cwc);
}

static int write_post_select(__a_unused struct sched *s, void *context)
{
	struct write_task *wt = context;
	return check_wav_post_select(wt->cwc);
}

static int setup_and_schedule(void)
{
	int i, ret;
	struct btr_node *cw_btrn;
	struct writer_node *wns;
	static struct sched s;
	struct wav_params wp;
	struct write_task wt;

	sit.btrn = btr_new_node(&(struct btr_node_description)
		EMBRACE(.name = "stdin"));
	stdin_task_register(&sit, &s);

	COPY_WAV_PARMS(&wp, &conf);
	wt.cwc = check_wav_init(sit.btrn, NULL, &wp, &cw_btrn);
	wt.task = task_register(&(struct task_info) {
		.name = "write",
		.pre_select = write_pre_select,
		.post_select = write_post_select,
		.context = &wt,
	}, &s);
	if (!conf.writer_given) {
		wns = para_calloc(sizeof(*wns));
		setup_writer_node(NULL, cw_btrn, wns, &s);
		i = 1;
	} else {
		wns = para_calloc(conf.writer_given * sizeof(*wns));
		for (i = 0; i < conf.writer_given; i++)
			setup_writer_node(conf.writer_arg[i], cw_btrn,
				wns + i, &s);
	}

	s.default_timeout.tv_sec = 10;
	s.default_timeout.tv_usec = 50000;
	ret = schedule(&s);
	if (ret >= 0) {
		int j, ts;
		for (j = 0; j < i; j++) {
			struct writer_node *wn = wns + j;
			ts = task_status(wn->task);
			assert(ts < 0);
			if (ts != -E_WRITE_COMMON_EOF && ts != -E_BTR_EOF) {
				const char *name = writer_names[wn->writer_num];
				PARA_ERROR_LOG("%s: %s\n", name,
					para_strerror(-ts));
				if (ret >= 0)
					ret = ts;
			}
		}
	}
	for (i--; i >= 0; i--) {
		struct writer_node *wn = wns + i;
		struct writer *w = writers + wn->writer_num;

		w->close(wn);
		btr_remove_node(&wn->btrn);
		w->free_config(wn->conf);
		free(wn->conf);
	}
	free(wns);
	check_wav_shutdown(wt.cwc);
	sched_shutdown(&s);
	return ret;
}

/**
 * Para_write's main function.
 *
 * \param argc The usual argument counter.
 * \param argv The usual argument vector.
 *
 * It sets up and starts the tasks and the buffer tree nodes determined by
 * command line options.
 *
 * \return \p EXIT_SUCCESS or EXIT_FAILURE
 */
int main(int argc, char *argv[])
{
	int ret;

	write_cmdline_parser(argc, argv, &conf);
	loglevel = get_loglevel_by_name(conf.loglevel_arg);
	writer_init();
	version_handle_flag("write", conf.version_given);
	if (conf.help_given || conf.detailed_help_given)
		print_help_and_die();

	ret = setup_and_schedule();
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		exit(EXIT_FAILURE);
	}
	exit(EXIT_SUCCESS);
}
